from secml_malware.attack.whitebox.c_end2end_evasion import CEnd2EndMalwareEvasion
from secml_malware.attack.whitebox.c_discretized_bytes_evasion import CDiscreteBytesEvasion
from secml_malware.attack.whitebox.c_fast_gradient_sign_evasion import CFastGradientSignMethodEvasion
from secml_malware.attack.whitebox.c_format_exploit_evasion import CFormatExploitEvasion
from secml_malware.attack.whitebox.c_header_evasion import CHeaderEvasion
from secml_malware.attack.whitebox.c_padding_evasion import CPaddingEvasion
from secml_malware.attack.whitebox.c_kreuk_evasion import CKreukEvasion
from secml_malware.attack.whitebox.c_suciu_evasion import CSuciuEvasion
from secml_malware.attack.whitebox.c_padding_evasion import CPaddingEvasion
from secml_malware.attack.whitebox.c_extend_dos_evasion import CExtendDOSEvasion
from secml_malware.attack.whitebox.c_shift_evasion import CContentShiftingEvasion