class FileNotExistsException(Exception):
	def __init__(self, message):
		super(FileNotExistsException, self).__init__(message)


class FolderNotExistsException(Exception):
	def __init__(self, message):
		super(FolderNotExistsException, self).__init__(message)

class OptimizerNotSetException(Exception):
	def __init__(self, message):
		super(OptimizerNotSetException, self).__init__(message)